@@ -21,16 +21,25 @@ import functools
 import json
 import os
 import mock
+import platform
 import re
 import socket
+import sys
 import unittest
 
 import pytest
 
-from pymemcache.client.base import PooledClient, Client, normalize_server_spec
+from pymemcache.client.base import (
+    PooledClient,
+    Client,
+    normalize_server_spec,
+    KeepaliveOpts,
+    check_key_helper
+)
 from pymemcache.exceptions import (
     MemcacheClientError,
     MemcacheServerError,
+    MemcacheUnexpectedCloseError,
     MemcacheUnknownCommandError,
     MemcacheUnknownError,
     MemcacheIllegalInputError
@@ -40,6 +49,49 @@ from pymemcache import pool
 from pymemcache.test.utils import MockMemcacheClient
 
 
+# TODO: Use ipaddress module when dropping support for Python < 3.3
+def is_ipv6(address):
+    return re.match(r'^[0-9a-f:]+$', address)
+
+
+@pytest.mark.parametrize(
+    'key,allow_unicode_keys,key_prefix,ex_exception,ex_excinfo,ignore_py27',
+    [
+        (u'b'*251, True, b'',
+         MemcacheIllegalInputError, 'Key is too long', False),
+        (u'foo bar', True, b'',
+         MemcacheIllegalInputError, 'Key contains whitespace', False),
+        (u'\00', True, b'',
+         MemcacheIllegalInputError, 'Key contains null', False),
+        (None, True, b'', TypeError, None, False),
+        # The following test won't fail with a TypeError with python 2.7
+        (b"", False, '', TypeError, None, True),
+    ])
+@pytest.mark.unit()
+def test_check_key_helper_failing_conditions(key, allow_unicode_keys,
+                                             key_prefix, ex_exception,
+                                             ex_excinfo, ignore_py27):
+
+    if ignore_py27 and sys.version_info < (3, 0, 0):
+        pytest.skip("skipping for Python 2.7")
+    with pytest.raises(ex_exception) as excinfo:
+        check_key_helper(key, allow_unicode_keys, key_prefix)
+
+    # Allow to ignore the excinfo value. Different implementation (2.7,
+    # pypy, 3.x) comes with various error messages for the same thing.
+    # We just check the kind of the exception.
+    if ex_excinfo:
+        assert ex_excinfo in str(excinfo.value)
+
+
+@pytest.mark.unit()
+def test_check_key_helper():
+    assert check_key_helper(b"key", True, b'') == b"key"
+    assert check_key_helper("key", True) == b"key"
+    assert isinstance(check_key_helper(b"key", True), bytes)
+    assert check_key_helper("", True) == b""
+
+
 class MockSocket(object):
     def __init__(self, recv_bufs, connect_failure=None, close_failure=None):
         self.recv_bufs = collections.deque(recv_bufs)
@@ -53,10 +105,8 @@ class MockSocket(object):
 
     @property
     def family(self):
-        # TODO: Use ipaddress module when dropping support for Python < 3.3
-        ipv6_re = re.compile(r'^[0-9a-f:]+$')
-        is_ipv6 = any(ipv6_re.match(c[0]) for c in self.connections)
-        return socket.AF_INET6 if is_ipv6 else socket.AF_INET
+        any_ipv6 = any(is_ipv6(c[0]) for c in self.connections)
+        return socket.AF_INET6 if any_ipv6 else socket.AF_INET
 
     def sendall(self, value):
         self.send_bufs.append(value)
@@ -115,6 +165,19 @@ class MockSocketModule(object):
         self.sockets.append(socket)
         return socket
 
+    def getaddrinfo(self, host, port, family=0, type=0, proto=0, flags=0):
+        family = family or (
+            socket.AF_INET6 if is_ipv6(host) else socket.AF_INET
+        )
+        type = type or socket.SOCK_STREAM
+        proto = proto or socket.IPPROTO_TCP
+        sockaddr = (
+            ('::1', 11211, 0, 0)
+            if family == socket.AF_INET6
+            else ('127.0.0.1', 11211)
+        )
+        return [(family, type, proto, '', sockaddr)]
+
     def __getattr__(self, name):
         return getattr(socket, name)
 
@@ -130,7 +193,7 @@ class CustomizedClient(Client):
 class ClientTestMixin(object):
     def make_client(self, mock_socket_values, **kwargs):
         client = Client(None, **kwargs)
-        # mock out client._connect() rather than hard-settting client.sock to
+        # mock out client._connect() rather than hard-setting client.sock to
         # ensure methods are checking whether self.sock is None before
         # attempting to use it
         sock = MockSocket(list(mock_socket_values))
@@ -140,7 +203,7 @@ class ClientTestMixin(object):
 
     def make_customized_client(self, mock_socket_values, **kwargs):
         client = CustomizedClient(None, **kwargs)
-        # mock out client._connect() rather than hard-settting client.sock to
+        # mock out client._connect() rather than hard-setting client.sock to
         # ensure methods are checking whether self.sock is None before
         # attempting to use it
         sock = MockSocket(list(mock_socket_values))
@@ -318,6 +381,16 @@ class ClientTestMixin(object):
         result = client.get(b'key', default='foobar')
         assert result == 'foobar'
 
+    def test_get_ignore_exc_default(self):
+        client = self.make_client([b'INVALID DATA\r\n'], ignore_exc=True)
+        result = client.get(b'key', default='foobar')
+        assert result == 'foobar'
+
+        # Unit test for customized client (override _extract_value)
+        client = self.make_client([b'INVALID DATA\r\n'], ignore_exc=True)
+        result = client.get(b'key', default='foobar')
+        assert result == 'foobar'
+
     def test_get_found(self):
         client = self.make_client([
             b'STORED\r\n',
@@ -672,6 +745,19 @@ class TestClient(ClientTestMixin, unittest.TestCase):
         assert result is None
         assert client.sock is None
 
+    def test_shutdown(self):
+        client = self.make_client([MemcacheUnexpectedCloseError('shutdown')])
+        result = client.shutdown()
+        assert result is None
+
+    def test_shutdown_disabled(self):
+        def _shutdown():
+            client = self.make_client([b'ERROR: shutdown not enabled\r\n'])
+            client.shutdown()
+
+        with pytest.raises(MemcacheUnknownCommandError):
+            _shutdown()
+
     def test_replace_stored(self):
         client = self.make_client([b'STORED\r\n'])
         result = client.replace(b'key', b'value', noreply=False)
@@ -784,6 +870,20 @@ class TestClient(ClientTestMixin, unittest.TestCase):
         with pytest.raises(MemcacheServerError):
             _set()
 
+    def test_closing_socket_on_unexpected_closed_error(self):
+        client = self.make_client([
+            b'VALUE ',
+            MemcacheUnexpectedCloseError("foo bar"),
+        ])
+
+        def _set():
+            client.set(b'key', b'value', noreply=False)
+
+        with pytest.raises(MemcacheUnexpectedCloseError):
+            _set()
+
+        assert client.sock is None
+
     def test_set_unknown_error(self):
         client = self.make_client([b'foobarbaz\r\n'])
 
@@ -1142,6 +1242,47 @@ class TestClientSocketConnect(unittest.TestCase):
             client._connect()
             assert client.sock.family == socket.AF_UNIX
 
+    @unittest.skipIf('Linux' != platform.system(),
+                     'Socket keepalive only support Linux platforms.')
+    def test_linux_socket_keepalive(self):
+        server = ('::1', 11211)
+        try:
+            client = Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=KeepaliveOpts())
+            client._connect()
+        except SystemError:
+            self.fail("SystemError unexpectedly raised")
+        with self.assertRaises(ValueError):
+            # A KeepaliveOpts object is expected, a ValueError will be raised
+            Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=True)
+
+    @mock.patch('platform.system')
+    def test_osx_socket_keepalive(self, platform_mock):
+        platform_mock.return_value = 'Darwin'
+        server = ('::1', 11211)
+        # For the moment the socket keepalive is only implemented for Linux
+        with self.assertRaises(SystemError):
+            Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=KeepaliveOpts())
+
+    @mock.patch('platform.system')
+    def test_windows_socket_keepalive(self, platform_mock):
+        platform_mock.return_value = 'Windows'
+        server = ('::1', 11211)
+        # For the moment the socket keepalive is only implemented for Linux
+        with self.assertRaises(SystemError):
+            Client(
+                server,
+                socket_module=MockSocketModule(),
+                socket_keepalive=KeepaliveOpts())
+
     def test_socket_connect_closes_on_failure(self):
         server = ("example.com", 11211)
 
@@ -1421,3 +1562,21 @@ class TestNormalizeServerSpec(unittest.TestCase):
         with pytest.raises(ValueError) as excinfo:
             f(12345)
         assert str(excinfo.value) == "Unknown server provided: 12345"
+
+
+@pytest.mark.unit()
+class TestKeepaliveopts(unittest.TestCase):
+    def test_keepalive_opts(self):
+        kao = KeepaliveOpts()
+        assert (kao.idle == 1 and kao.intvl == 1 and kao.cnt == 5)
+        kao = KeepaliveOpts(idle=1, intvl=4, cnt=2)
+        assert (kao.idle == 1 and kao.intvl == 4 and kao.cnt == 2)
+        kao = KeepaliveOpts(idle=8)
+        assert (kao.idle == 8 and kao.intvl == 1 and kao.cnt == 5)
+        kao = KeepaliveOpts(cnt=8)
+        assert (kao.idle == 1 and kao.intvl == 1 and kao.cnt == 8)
+
+        with self.assertRaises(ValueError):
+            KeepaliveOpts(cnt=0)
+        with self.assertRaises(ValueError):
+            KeepaliveOpts(idle=-1)
